# -*- coding: cp1252 -*-
# (C) by Fred Bayer <fred@bayerf.de>
#
# Sun rise/set for the F4LW
# Using the algorithm from
# http://lexikon.astronomie.info/zeitgleichung/neu.html
#
# Also computes (simplified) moon phase and assembled RTC data for F4LW

from math import *
import time

pi2 = 2*pi

synMonth = 29.530589  # 29d 12h 44m 2.9s

def simpleMoon(y, m, d):
  """Compute half days modulo 59, just like the F4LW using basic arithmetic only
     (see label MoonPhase in f4lw.asm)."""
  r = 0
  mlen = [0, 3, 0, 3, 4, 7, 8, 11, 14, 15, 18, 19] # cumulated halfdays mod 59

  y%=100                            # F4LW uses 2 digit year only
  y4 = y>>2
  while y4 > 0:
    r += 31 # (365*3+366)*2 % 59    # halfdays in 3 years + 1 leapyear
    if r&0x80:                      # when too large, normalize
      r -= 59
    y4 -= 1
  y &= 3

  if not y and m<=2:                # adjust for first two months of leap year
    r -= 2
  while y > 0:
    r += 22  # 365*2 % 59           # r <= 127 + 3*22 = 193, no danger of overflow
    y -= 1

  r += mlen[m-1]                    # r <= 193 + 19 = 212, no overflow
  if r&0x80:                        # but normalize now
    r -= 59
  r += 2*d

  while r>=59:                      # final nomalization
    r -= 59
  return r

def correctMoon(y, m ,d):
  "Compute moon age in synodic months."
  return fmod((jDate(y, m, d)-newMoon)/synMonth,1)

def moonGauge(y, m, d):
  "Compute the initial moon age for the simplified calculation in the F4lW."
  return (int(correctMoon(y,m,d)*59+0.5) - simpleMoon(y,m,d) + 2) % 59

def jDate(y, mo, d, h=12, mn=0, s=0):
  "Compute Julian day number from date components."
  if mo<=2:
    mo+=12
    y-=1
  g = y//400-y//100+y//4
  return 2400000.5+365*y - 679004 + g +\
         int(30.6001*(mo+1)) + d + h/24.0 +\
         mn/1440.0 + s/86400.0

newMoon = jDate(2000, 1, 6, 19, 13.6)

def eps(t):
  "Inflection of earth axis."
  return radians(23.43929111 + (-46.815*t - 0.00059*t*t + 0.001813*t*t*t)/3600)

def eqTime(t):
  "Compute equation of time."
  def red(a):
    res = fmod(a,pi2)
    if res<0: res += pi2
    return res

  raAv = 18.71506921 + 2400.0513369*t + (2.5862e-5 - 1.72e-9*t)*t*t
  m = red(pi2 * (0.993133 + 99.997361*t))
  l = red(pi2 * (0.7859453 + m/pi2 + (6893.0*sin(m)+72*sin(2*m)+6191.2*t)/1296e3))
  e = eps(t)
  ra = atan(tan(l)*cos(e))
  if ra<0: ra += pi
  if l>pi: ra += pi

  ra = 24*ra/pi2
  dk = asin(sin(e)*sin(l))

  # Normalise to 0<=raAv<24
  raAv = fmod(raAv, 24.0)
  if raAv<0 : raAv += 24.0

  dra = raAv - ra
  if dra<-12: dra += 24
  if dra>12:  dra -= 24

  dra *= 1.0027379

  return dra, dk

def mkDate(y,m,d):
  "Create time_struct with DST flag set correctly."
  return time.localtime(time.mktime((y,m,d,12,0,0,0,0,-1)))

def rs(dt, latitude, longitude):
  "Compute sun rise/set for given date and coordinates."
  tz = time.timezone/-3600.0 # Care for DST in F4LW
  jd2000 = 2451545.0
  jd = jDate(dt[0], dt[1], dt[2])
  t = (jd-jd2000)/36525
  h = radians(-50.0/60.0)
  b = radians(latitude)

  eqt, dk = eqTime(t)
  dt = 12*acos((sin(h)-sin(b)*sin(dk)) / (cos(b)*cos(dk)))/pi

  srise = 12 - dt - eqt - longitude/15.0 + tz
  sset  = 12 + dt - eqt - longitude/15.0 + tz

  return srise, sset

import array
def makeSunblock(year, latitude, longitude):
  "Compute sun rise/set times for the entire year and store them in BCD format."
  def bcd(n):
    return int(str(n),16)
  mem = array.array("B", " "*(23*64))
  NA  = array.array("B", [0xff]*4) # No rise/set at this date
  for d in range(366): 
    date = mkDate(year,1,d+1) # mkDate cares for normalization!
    adr = 4*d
    try:
      r,s = rs(date, latitude, longitude)
      rh = int(r)
      rm = max(0,int(fmod(r,1)*60+0.5))
      if rm>=60: rm=0; rh+=1
      sh = int(s)
      sm = max(0,int(fmod(s,1)*60+0.5))
      if sm>=60: sm=0; sh+=1
      if sh>=24: raise OverflowError
      mem[adr]   = bcd(rh)
      mem[adr+1] = bcd(rm)
      mem[adr+2] = bcd(sh)
      mem[adr+3] = bcd(sm)
    except (ValueError,OverflowError):
      mem[adr:adr+4] = NA
  return mem

def timeAsRtc(tm, includeSun=False):
  "Return current time and date as a binary string for the DS1307 RTC."

  # Compute gauge value for moonphase calculation
  moon = moonGauge(tm.tm_year, tm.tm_mon, tm.tm_mday)

  # tm_wday is 0..6 (0 is Monday), but RTC expects 1..7 (1 is Monday)
  # strftime specifier %w returns 0..6 with 0 is Sunday, so
  # we have to to it manually :-(
  # The 7th byte is the control byte enabling square wave @ 1Hz,
  # the 8th byte is the century. Remember to fix in 93 years! 	
  # the 9th byte is the daylight saving time flag (bit 0) and a
  #              flag (bit 7) if the sun table will follow
  # the 10th byte is the new-moon offset in half days (not BCD)	
  return (time.strftime("%S%M%H0%%d%d%m%y1020%%02x%%02x", tm) %\
          (tm.tm_wday+1, tm.tm_isdst+128*includeSun, moon)).decode("hex")

  
  
