# -*- coding: cp1252 -*-
# (C) by Fred Bayer <fred@bayerf.de>
#

"""\
Some word experiments: Group word lists from the file <testdic> (one word
per line) into sets of equal length. Build chains (class Chain) of words of
length <n> which differ at atmost <mmm> positions and provide a generator
to generate these word chains. Also, assemble the memory image for the
F4LW linguistic sculpture to do the same.
"""

import sys
import time
import random
import os
from collections import deque
import array
import re

#filename, total # words, # of 4-letter words, total length of Levenstein-1 chains
#testdic = "3esl.txt" # 21877/1568/10462
#testdic = "6of12.txt" # 32153/1822/13070
#testdic = "2of12.txt" # 41238/2024/15476
#testdic = "2of12inf.txt" # 81520/2520/22446
testdic = "5desk.txt" # 61406/2731/21644
#testdic = "2of4brif.txt" # 60387/2360/20112
#testdic = "..\\word.list" # 264061/5248/68130
#testdic = "..\\germanbig.txt" 
#testdic = "..\\cycletest.txt" 

testdic = r"..\words\12dict\\"+testdic

# Bug alert: define special isalpha(), upper() since
# locale.setlocale(locale.LC_ALL,"") (which should make umlauts work)
# makes Backspace key non-functional in Tk widgets !?!?!

def myIsalpha(s):
  for c in s:
    if not c.isalpha() and c not in "":
      return False
  return True

def myUpper(s):
  return s.replace("","#").upper().replace("","").replace("","").replace("","").\
         replace("","").replace("","").replace("#","")

# The mapping for F4LW umlauts and accented chars
umltab=list(chr(n) for n in range(0,128)) + [" "]*128
umltab[ord("")]="["
umltab[ord("")]="\\"
umltab[ord("")]="]"
umltab[ord("")]="{"
umltab[ord("")]="|"
umltab[ord("")]="}"
umltab[ord("")]="\x02"
umltab[ord("")]="A"
umltab[ord("")]="E"
umltab[ord("")]="a"
umltab[ord("")]="e"
langTrans = {"en": "".join(chr(n) for n in range(0,128)) + " "*128,
             "de": "".join(umltab)}

def setWordlist(filename):
  global words, groups
  words = []
  for line in file(filename):
    words.append(myUpper(line[:-1]))
  groups = makeGroups()

def makeGroups():
  gr = {}
  for w in words:
    l = len(w)
    if l in gr:
      gr[l].append(w)
    else:
      gr[l] = [w]
  for i in gr:
    gr[i] = list(set(w for w in gr[i] if myIsalpha(w)))
  return gr

def sizes(l=range(0,50)):
  """Print the number of words in the dictionary for each word length in <l> and
  the total size of the dict in bytes."""
  sum = 0
  for i in l:
    g = groups.get(i)
    if g:
      sum += i*len(g)
      print i,len(g)
  print "total",sum

class Chain:
  def __init__(self, n, mmm=1):
    self.n = n
    self.mmm = mmm
    self.words = groups[n]
    self.graph = dict(((w,set()) for w in self.words))
    self.range = range(self.n)

    def checkmismatch(a,b):
      mm=0
      for i in self.range:
        if a[i]!=b[i]:
          mm +=1
          if mm>mmm:
            return False
      return mm>0
    
    def add(tw):
      for w in self.words:
        if checkmismatch(w, tw):
          self.graph[tw].add(w)

    for w in self.words:
      add(w)

  def __len__(self):
    return len(self.words)

  def __repr__(self):
    return "Chain(len=%d, max mismatch=%d, %d words, %d chains)" %\
           (self.n, self.mmm, len(self), sum(len(self.graph[p]) for p in self.words))

  def cliques(self):
    """Split the word graph into connected components."""
    visited = set()
    cls = []
    for w in self.words:
      if w not in visited:
        cl = set()
        op = deque()
        op.append(w)
        while len(op):
          w1 = op.popleft()
          for w2 in self.graph[w1]:
            if w2 not in visited:
              op.append(w2)
          cl.add(w1)
          visited.add(w1)
        cls.append(cl)
    return cls
        

  def rw(self, hsize=1):
    """Generate random words differing by atmost <mmm> positions. If a dead end
    is encountered, None is returned and the next iteration starts with a fresh
    random word. <norepeats> is the size of the word history to be excluded to
    avoid early repetition. Normally, this generator runs forever, but if it is
    unable to provide a word not yet in the history (the history is larger than
    the word database), it will stop.
    """
    history = deque()

    def selword(wset):
      rset = wset-set(history)
      if rset:
        word = random.choice(list(rset))
        if len(history) >= hsize:
          history.popleft()
        history.append(word)
        return word
      else:
        return None
      
    while True:
      word = selword(set(self.words))
      if not word:
        raise StopIteration()
      yield word
      while True:
        word = selword(self.graph[word])
        yield word
        if not word:
          break

  def makemem(self, tag, lang="en"):
    """Assemble a memory image suitable for download to the F4LW linguistic
    sculpture. Layout is:
      * 1 byte eyecatcher 0x55
      * 2 bytes number of words
      * 4 bytes tag
      * for each word:
        * 2 byte pointer to transition table entry
        * word in uppercase
      * transition table entries
        * 1 byte length, highest bit set if word is member of a small clique,
          so chains may be broken here to avoid displaying same group forever
        * indicated no. pointers.
    All data items are packed (aligned on byte boundaries). 2 byte values are
    stored in big-endian format.
    """
    assert len(tag)==4
    nw = len(self.words)
    size = 7 + nw*(self.n+2) + sum(len(self.graph[w])*2+1 for w in self.words)
    if size>0xfa00:  # Upper 1536 bytes used for sunrise/set table
      raise ValueError("Size %d too large" % size)

    cliques = self.cliques()
    def inSmallClique(w):
      for c in cliques:
        if w in c:
          return 2<len(c)<10
      assert False, "I cannot be here!"
      
    mem = array.array("B", [0]*size)
    mem[0] = 0x55
    mem[1] = nw//256
    mem[2] = nw%256
    for i in range(4):
      mem[3+i] = ord(tag[i])
    addr_w = 7
    addr_c = 7 + nw*(self.n+2)
    startpos = dict()
    for w in self.words:
      startpos[w] = addr_w
      for i in range(self.n):
        mem[addr_w+i+2] = ord(w[i].translate(langTrans[lang]))
      addr_w += self.n+2
    assert addr_w==addr_c
    for w,p in startpos.iteritems():
      mem[p] = addr_c//256
      mem[p+1] = addr_c%256
      wset=self.graph[w]
      wsize=len(wset)
      assert wsize < 128
      if inSmallClique(w):
        wsize |= 0x80
      mem[addr_c] = wsize
      addr_c += 1
      for cw in wset:
        p = startpos[cw]
        mem[addr_c] = p//256
        mem[addr_c+1] = p%256
        addr_c += 2
    assert addr_c==size
    return mem

  def savemem(self, filename, *args):
    self.makemem(*args).tofile(file(filename, "wb"))
